/**
 * Annotations that directly depend on classes in databinding bundle
 * (not just Jackson core) and can not be included
 * in Jackson core annotations package (because it can not have any
 * external dependencies).
 */
package com.fasterxml.jackson.databind.annotation;
