/**
 * Contains concrete {@link com.fasterxml.jackson.databind.JsonNode} implementations
 * Jackson uses for the Tree model.
 * These classes are public since concrete type will be needed
 * for most operations that modify node trees. For read-only access concrete
 * types are usually not needed.
 */
package com.fasterxml.jackson.databind.node;
