/**
Package that contains most of configuration-related classes;
exception being couple of most-commonly used configuration
things (like Feature enumerations) that are at the
main level (<code>com.fasterxml.jackson.databind</code>).
*/

package com.fasterxml.jackson.databind.cfg;
