/**
 * Contains those implementation classes of deserialization part of 
 * data binding that are not considered part of public or semi-public
 * interfaces. Use of these classes by non-core classes is discouraged,
 * although occasionally this may be necessary.
 * Note that backwards-compatibility of these classes is not guaranteed
 * between minor releases (but is between patch releases).
 */
package com.fasterxml.jackson.databind.deser.impl;
